from django.conf import settings
from django.contrib.auth.models import Permission
from django.urls import path, re_path, reverse
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.users.widgets import UserListingButton

from wagtail_2fa import views


@hooks.register("register_admin_urls")
def urlpatterns():
    return [
        path("2fa/auth", views.LoginView.as_view(), name="wagtail_2fa_auth"),
        re_path(
            r"^2fa/devices/(?P<user_id>\d+)$",
            views.DeviceListView.as_view(),
            name="wagtail_2fa_device_list",
        ),
        path(
            "2fa/devices/new",
            views.DeviceCreateView.as_view(),
            name="wagtail_2fa_device_new",
        ),
        re_path(
            r"^2fa/devices/(?P<pk>\d+)/update$",
            views.DeviceUpdateView.as_view(),
            name="wagtail_2fa_device_update",
        ),
        re_path(
            r"^2fa/devices/(?P<pk>\d+)/remove$",
            views.DeviceDeleteView.as_view(),
            name="wagtail_2fa_device_remove",
        ),
        re_path(
            r"^2fa/devices/qr-code$",
            views.DeviceQRCodeView.as_view(),
            name="wagtail_2fa_device_qrcode",
        ),
    ]


@hooks.register("construct_main_menu")
def remove_menu_if_unverified(request, menu_items):
    """Remove the sidebar menu items if the user is unverified."""
    if (
        settings.WAGTAIL_2FA_REQUIRED
        and getattr(request.user, "enable_2fa", True)
        and not request.user.is_verified()
    ):
        menu_items.clear()
        menu_items.append(
            MenuItem(
                "2FA Setup",
                reverse("wagtail_2fa_device_list", kwargs={"user_id": request.user.id}),
            )
        )


@hooks.register("register_account_menu_item")
def register(request):
    return {
        "url": reverse("wagtail_2fa_device_list", kwargs={"user_id": request.user.id}),
        "label": _("Manage your 2FA devices"),
        "help_text": _("Add or remove devices for 2 factor authentication."),
    }


@hooks.register("register_user_listing_buttons")
def register_user_listing_buttons(context, user):
    yield UserListingButton(
        _("Manage 2FA"),
        reverse("wagtail_2fa_device_list", kwargs={"user_id": user.id}),
        attrs={"title": _("Edit this user")},
        priority=100,
    )


@hooks.register("register_permissions")
def register_2fa_permission():
    if "wagtail_2fa.middleware.VerifyUserPermissionsMiddleware" in settings.MIDDLEWARE:
        return Permission.objects.filter(
            content_type__app_label="wagtailadmin", codename="enable_2fa"
        )

    return Permission.objects.none()



from django.utils.translation import gettext_lazy as _
from django_otp.plugins.otp_totp.models import TOTPDevice


def get_unconfirmed_device(user):
    return TOTPDevice.objects.devices_for_user(user, confirmed=False).first()


def new_unconfirmed_device(user):
    delete_unconfirmed_devices(user)
    num = TOTPDevice.objects.filter(user=user).count()
    return TOTPDevice.objects.create(
        name=_("Device #%s") % (num + 1), user=user, confirmed=False
    )


def delete_unconfirmed_devices(user):
    (TOTPDevice.objects.devices_for_user(user, confirmed=False).delete())



from functools import partial

import django_otp
from django.conf import settings
from django.contrib.auth.views import redirect_to_login
from django.urls import resolve, reverse
from django.utils.functional import SimpleLazyObject
from django_otp.middleware import OTPMiddleware as _OTPMiddleware


class VerifyUserMiddleware(_OTPMiddleware):
    _allowed_url_names = [
        "wagtail_2fa_auth",
        "wagtailadmin_login",
        "wagtailadmin_logout",
        "wagtailadmin_javascript_catalog",
        "wagtailadmin_sprite",
    ]

    # These URLs do not require verification if the user has no devices
    _allowed_url_names_no_device = [
        "wagtail_2fa_device_list",
        "wagtail_2fa_device_new",
        "wagtail_2fa_device_qrcode",
    ]

    def __call__(self, request):
        if hasattr(self, "process_request"):
            response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        if hasattr(self, "process_response"):
            response = self.process_response(request, response)
        return response

    def process_request(self, request):
        if request.user:
            request.user = SimpleLazyObject(
                partial(self._verify_user, request, request.user)
            )
        user = request.user
        if self._require_verified_user(request):
            user_has_device = django_otp.user_has_device(user, confirmed=True)

            if user_has_device and not user.is_verified():
                return redirect_to_login(
                    request.get_full_path(), login_url=reverse("wagtail_2fa_auth")
                )

            elif not user_has_device and settings.WAGTAIL_2FA_REQUIRED:
                # only allow the user to visit the admin index page and the
                # admin setup page
                return redirect_to_login(
                    request.get_full_path(), login_url=reverse("wagtail_2fa_device_new")
                )

    def _require_verified_user(self, request):
        user = request.user

        if not settings.WAGTAIL_2FA_REQUIRED:
            # If two factor authentication is disabled in the settings
            return False

        if not user.is_authenticated:
            return False

        # If the user has no access to the admin anyway then don't require a
        # verified user here
        if not (
            user.is_staff
            or user.is_superuser
            or user.has_perms(["wagtailadmin.access_admin"])
        ):
            return False

        # Don't require verification for specified URL names
        request_url_name = resolve(request.path_info).url_name
        if request_url_name in self._allowed_url_names:
            return False

        # If the user does not have a device, don't require verification
        # for the specified URL names
        if request_url_name in self._allowed_url_names_no_device:
            user_has_device = django_otp.user_has_device(user, confirmed=True)
            if not user_has_device:
                return False

        # For all other cases require that the user is verified via otp
        return True


class VerifyUserPermissionsMiddleware(VerifyUserMiddleware):
    """A variant of VerifyUserMiddleware which makes 2FA optional."""

    def process_request(self, request):
        result = super().process_request(request)

        # Add an attribute to the user so we can easily determine if 2FA should
        # be enabled for them.
        request.user.enable_2fa = request.user.has_perms(["wagtailadmin.enable_2fa"])

        return result

    def _require_verified_user(self, request):
        result = super()._require_verified_user(request)

        # Always require verification if the user has a device, even if they have
        # 2FA disabled.
        user_has_device = django_otp.user_has_device(request.user, confirmed=True)
        if not user_has_device and not request.user.has_perms(
            ["wagtailadmin.enable_2fa"]
        ):
            return False

        return result



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django_otp import user_has_device


class OtpRequiredMixin(object):
    """
    if_configured: If ``True``, an authenticated user with no confirmed
        OTP devices will be allowed. Default is ``False``.
    """

    if_configured = False

    def handle_no_permission(self, request):
        """Redirect unauthenticated users."""
        if not request.user.is_authenticated:
            return redirect_to_login(
                request.get_full_path(), settings.LOGIN_URL, REDIRECT_FIELD_NAME
            )

        if user_has_device(request.user):
            return redirect_to_login(
                request.get_full_path(), login_url=reverse("wagtail_2fa_auth")
            )

        raise PermissionDenied

    def user_allowed(self, user):
        if not settings.WAGTAIL_2FA_REQUIRED:
            return True

        return user.is_verified() or (
            self.if_configured and user.is_authenticated and not user_has_device(user)
        )

    def dispatch(self, request, *args, **kwargs):
        if not self.user_allowed(request.user):
            return self.handle_no_permission(request)

        return super(OtpRequiredMixin, self).dispatch(request, *args, **kwargs)



from django.apps import AppConfig
from django.conf import settings

WAGTAIL_2FA_DEFAULT_SETTINGS = {
    "WAGTAIL_2FA_REQUIRED": False,
    "WAGTAIL_2FA_OTP_TOTP_NAME": False,
}


class Wagtail2faConfig(AppConfig):
    name = "wagtail_2fa"
    label = "wagtail_2fa"
    verbose_name = "Wagtail 2FA"

    def ready(self):
        for setting_name, setting_default_value in WAGTAIL_2FA_DEFAULT_SETTINGS.items():
            if not hasattr(settings, setting_name):
                setattr(settings, setting_name, setting_default_value)

        # Set OTP_TOTP_ISSUER, to identify the app, in authenticator
        # only set this if it's not already set
        if not hasattr(settings, "OTP_TOTP_ISSUER"):
            if not settings.WAGTAIL_2FA_OTP_TOTP_NAME and hasattr(
                settings, "WAGTAIL_SITE_NAME"
            ):
                settings.OTP_TOTP_ISSUER = settings.WAGTAIL_SITE_NAME
            elif settings.WAGTAIL_2FA_OTP_TOTP_NAME:
                settings.OTP_TOTP_ISSUER = settings.WAGTAIL_2FA_OTP_TOTP_NAME



from django import forms
from django.contrib.auth import authenticate
from django.utils.translation import gettext_lazy as _
from django_otp.forms import OTPAuthenticationFormMixin
from django_otp.plugins.otp_totp.models import TOTPDevice


class TokenForm(OTPAuthenticationFormMixin, forms.Form):
    otp_token = forms.CharField(required=True)

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super().__init__(*args, **kwargs)

        self.fields["otp_token"].widget.attrs.update(
            {"autofocus": "autofocus", "autocomplete": "off"}
        )

    def clean(self):
        self.clean_otp(self.user)
        return self.cleaned_data


class DeviceForm(forms.ModelForm):
    otp_token = forms.CharField(
        label=_("OTP token"),
        required=True,
        help_text=_(
            "Enter the numeric code displayed on your device after scanning the QR code"
        ),
    )

    name = forms.CharField(
        label=_("Name"),
        required=True,
        help_text=_("The human-readable name of this device."),
    )

    password = forms.CharField(
        label=_("Current password"),
        help_text=_("As an extra security measure, we need your current password."),
        widget=forms.PasswordInput(),
    )

    class Meta:
        model = TOTPDevice
        fields = ["name", "otp_token"]

    def __init__(self, request, **kwargs):
        super().__init__(**kwargs)
        self.fields["otp_token"].widget.attrs.update(
            {"autofocus": "autofocus", "autocomplete": "off"}
        )
        if self.instance.confirmed:
            del self.fields["otp_token"]

        self.request = request

    def clean_otp_token(self):
        token = self.cleaned_data.get("otp_token")

        if token and self.instance.verify_token(token):
            return token

        raise forms.ValidationError(
            _("Invalid token. Please make sure you have entered it correctly."),
            code="invalid",
        )

    def clean_password(self):
        password = self.cleaned_data.get("password")
        if not authenticate(
            self.request, username=self.request.user.get_username(), password=password
        ):
            raise forms.ValidationError(_("Invalid password"))

    def save(self):
        self.instance.confirmed = True
        self.instance.save()



import re

from setuptools import find_packages, setup

install_requires = [
    "Django>=3.2",
    "Wagtail>=4.1",
    "django-otp>=0.8.1",
    "six>=1.14.0",
    "qrcode>=6.1",
]

docs_require = [
    "sphinx>=1.4.1",
    "sphinx_rtd_theme>=0.4.3",
]

tests_require = [
    "coverage==5.5",
    "pytest==6.2.4",
    "pytest-cov==2.12.1",
    "pytest-django==4.4.0",
    # Linting
    "flake8==3.9.2",  # 3.7.9
    "isort==5.9.3",
    "flake8-blind-except==0.2.0",
    "flake8-debugger==4.0.0",
]

with open("README.rst") as fh:
    long_description = re.sub(
        "^.. start-no-pypi.*^.. end-no-pypi", "", fh.read(), flags=re.M | re.S
    )

setup(
    name="wagtail-2fa",
    version="1.6.7",
    description="Two factor authentication for Wagtail",
    long_description=long_description,
    url="https://github.com/LabD/wagtail-2fa",
    author="Lab Digital",
    author_email="opensource@labdigital.nl",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        "docs": docs_require,
        "test": tests_require,
    },
    python_requires=">=3.8",
    use_scm_version=True,
    entry_points={},
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Framework :: Wagtail",
        "Framework :: Wagtail :: 2",
        "Framework :: Wagtail :: 3",
        "Framework :: Wagtail :: 4",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    zip_safe=False,
)



import qrcode
import qrcode.image.svg
from django import VERSION as DJANGO_VERSION
from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME

if DJANGO_VERSION >= (4, 1):
    from django.contrib.auth.views import RedirectURLMixin
else:
    from django.contrib.auth.views import SuccessURLAllowedHostsMixin as RedirectURLMixin

from wagtail import VERSION as WAGTAIL_VERSION

from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.http import url_has_allowed_host_and_scheme
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic import (
    DeleteView, FormView, ListView, UpdateView, View)
from django_otp import login as otp_login
from django_otp.plugins.otp_totp.models import TOTPDevice

from wagtail_2fa import forms, utils
from wagtail_2fa.mixins import OtpRequiredMixin


class LoginView(RedirectURLMixin, FormView):

    if WAGTAIL_VERSION >= (5, 0):
        template_name = "wagtail_2fa/otp_form.html"
    else:
        template_name = "wagtail_2fa/legacy/otp_form.html"

    form_class = forms.TokenForm
    redirect_field_name = REDIRECT_FIELD_NAME

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context[self.redirect_field_name] = self.get_redirect_url()
        return context

    def form_valid(self, form):
        otp_login(self.request, self.request.user.otp_device)
        return super().form_valid(form)

    def get_redirect_url(self):
        """Return the user-originating redirect URL if it's safe."""
        redirect_to = self.request.POST.get(
            self.redirect_field_name, self.request.GET.get(self.redirect_field_name, "")
        )
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=self.get_success_url_allowed_hosts(),
            require_https=self.request.is_secure(),
        )
        return redirect_to if url_is_safe else ""

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url(settings.LOGIN_REDIRECT_URL)


class DeviceListView(OtpRequiredMixin, ListView):
    template_name = "wagtail_2fa/device_list.html"

    # require OTP if configured
    if_configured = True

    def get_queryset(self):
        return TOTPDevice.objects.devices_for_user(
            self.kwargs["user_id"], confirmed=True
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["user_id"] = int(self.kwargs["user_id"])
        return context

    def dispatch(self, request, *args, **kwargs):
        """Override dispatch to ensure user is allowed to list the devices.

        Users are always allowed to list their own 2FA devices.
        If they have the ``change_user`` permission, they are allowed
        to list other users' 2FA devices.

        """
        if int(self.kwargs["user_id"]) == request.user.pk or request.user.has_perm(
            "user.change_user"
        ):
            if not self.user_allowed(request.user):
                return self.handle_no_permission(request)

            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied


class DeviceCreateView(OtpRequiredMixin, FormView):
    form_class = forms.DeviceForm
    template_name = "wagtail_2fa/device_form.html"

    # require OTP if configured
    if_configured = True

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["request"] = self.request
        kwargs["instance"] = self.device
        return kwargs

    def form_valid(self, form):
        form.save()
        utils.delete_unconfirmed_devices(self.request.user)

        if not self.request.user.is_verified():
            otp_login(self.request, form.instance)
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "wagtail_2fa_device_list", kwargs={"user_id": self.request.user.id}
        )

    @cached_property
    def device(self):
        if self.request.method.lower() == "get":
            return utils.new_unconfirmed_device(self.request.user)
        else:
            return utils.get_unconfirmed_device(self.request.user)


class DeviceUpdateView(OtpRequiredMixin, UpdateView):
    form_class = forms.DeviceForm
    template_name = "wagtail_2fa/device_form.html"

    def get_queryset(self):
        return TOTPDevice.objects.devices_for_user(self.request.user, confirmed=True)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["request"] = self.request
        return kwargs

    def get_success_url(self):
        return reverse(
            "wagtail_2fa_device_list", kwargs={"user_id": self.request.user.id}
        )


class DeviceDeleteView(OtpRequiredMixin, DeleteView):
    template_name = "wagtail_2fa/device_confirm_delete.html"

    def get_queryset(self):
        device = TOTPDevice.objects.get(**self.kwargs)
        return TOTPDevice.objects.devices_for_user(device.user, confirmed=True)

    def get_success_url(self):
        return reverse(
            "wagtail_2fa_device_list",
            kwargs={"user_id": self.request.POST.get("user_id")},
        )

    def dispatch(self, request, *args, **kwargs):
        """Override dispatch to ensure user is allowed to delete the device.

        Users are always allowed to delete their own 2FA devices.
        If they have the ``change_user`` permission, they are allowed
        to delete other users' 2FA devices.
        """
        device = TOTPDevice.objects.get(**self.kwargs)
        if device.user.pk == request.user.pk or request.user.has_perm(
            "user.change_user"
        ):
            if not self.user_allowed(request.user):
                return self.handle_no_permission(request)

            return super().dispatch(request, *args, **kwargs)

        raise PermissionDenied


class DeviceQRCodeView(OtpRequiredMixin, View):
    # require OTP if configured
    if_configured = True

    def get(self, request):
        device = utils.get_unconfirmed_device(self.request.user)
        img = qrcode.make(device.config_url, image_factory=qrcode.image.svg.SvgImage)
        response = HttpResponse(content_type="image/svg+xml")
        img.save(response)

        return response



# Generated by Django 2.2.7 on 2019-11-14 15:01

from django.db import migrations


def create_2fa_permissions(apps, schema_editor):
    ContentType = apps.get_model('contenttypes.ContentType')
    Permission = apps.get_model('auth.Permission')

    wagtailadmin_content_type, created = ContentType.objects.get_or_create(
        app_label='wagtailadmin',
        model='admin'
    )

    # Create 2FA permission
    enable_2fa_permission, created = Permission.objects.get_or_create(
        content_type=wagtailadmin_content_type,
        codename='enable_2fa',
        name='Enable 2FA'
    )


def remove_2fa_permissions(apps, schema_editor):
    """Reverse the above additions of permissions."""
    ContentType = apps.get_model('contenttypes.ContentType')
    Permission = apps.get_model('auth.Permission')
    wagtailadmin_content_type = ContentType.objects.get(
        app_label='wagtailadmin',
        model='admin',
    )

    # This also removes the permission from all groups
    Permission.objects.filter(
        content_type=wagtailadmin_content_type,
        codename='enable_2fa',
    ).delete()


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.RunPython(create_2fa_permissions, remove_2fa_permissions),
    ]
